from itertools import repeat
from typing import Sequence, Union

from torch_geometric.data import Data
from torch_geometric.data.datapipes import functional_transform
from torch_geometric.transforms import BaseTransform


@functional_transform('random_jitter')
class RandomJitter(BaseTransform):
    r"""Translates node positions by randomly sampled translation values
    within a given interval (functional name: :obj:`random_jitter`).
    In contrast to other random transformations,
    translation is applied separately at each position.

    Args:
        translate (sequence or float or int): Maximum translation in each
            dimension, defining the range
            :math:`(-\mathrm{translate}, +\mathrm{translate})` to sample from.
            If :obj:`translate` is a number instead of a sequence, the same
            range is used for each dimension.
    """
    def __init__(
        self,
        translate: Union[float, int, Sequence[Union[float, int]]],
    ) -> None:
        self.translate = translate

    def forward(self, data: Data) -> Data:
        assert data.pos is not None
        num_nodes, dim = data.pos.size()

        translate: Sequence[Union[float, int]]
        if isinstance(self.translate, (int, float)):
            translate = list(repeat(self.translate, times=dim))
        else:
            assert len(self.translate) == dim
            translate = self.translate

        jitter = data.pos.new_empty(num_nodes, dim)
        for d in range(dim):
            jitter[:, d].uniform_(-abs(translate[d]), abs(translate[d]))

        data.pos = data.pos + jitter

        return data

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.translate})'



from typing import Any

import torch
from torch import Tensor
from torch.autograd import grad


def influence(model: torch.nn.Module, src: Tensor, *args: Any) -> Tensor:
    x = src.clone().requires_grad_()
    out = model(x, *args).sum(dim=-1)

    influences = []
    for j in range(src.size(0)):
        influence = grad([out[j]], [x], retain_graph=True)[0].abs().sum(dim=-1)
        influences.append(influence / influence.sum())

    return torch.stack(influences, dim=0)



from math import isnan
from typing import Optional

from torch_geometric.llm.models.txt2kg import \
    _chunk_to_triples_str_cloud as call_NIM

# Credit for original "Marlin Accuracy" system goes to:
# Gilberto Titericz (NVIDIA)
# This work is an adaptation of his for PyG
SYSTEM_PROMPT_1 = (
    "Instruction: You are a world class state of the art " +
    "assistant for rating " +
    "a User Answer given a Question. The Question is completely" +
    " answered by the Reference Answer.\n" +
    "Say 4, if User Answer is full contained and equivalent to" +
    " Reference Answer" +
    "in all terms, topics, numbers, metrics, dates and units.\n" +
    "Say 2, if User Answer is partially contained and almost " +
    "equivalent to Reference Answer" +
    "in all terms, topics, numbers, metrics, dates and units.\n" +
    "Say 0, if User Answer is not contained in Reference Answer" +
    " or not accurate in all terms, topics," +
    "numbers, metrics, dates and units or the User Answer do not" +
    " answer the question.\n" +
    "Do not explain or justify your rating. Your rating must be " +
    "only 4, 2 or 0 according to the instructions above.\n" +
    "### Question: \"{question}\"\n" + "### User Answer: \"{model_pred}\"\n" +
    "### Reference Answer: \"{correct_answer}\"\n" + "The rating is:\n")

SYSTEM_PROMPT_2 = (
    "I will rate the User Answer in comparison to the Reference " +
    "Answer for a given Question.\n" +
    "A rating of 4 indicates that the User Answer is entirely " +
    "consistent with the Reference Answer, covering all aspects," +
    " topics, numbers, metrics, dates, and units.\n" +
    "A rating of 2 signifies that the User Answer is mostly " +
    "aligned with the Reference Answer, with minor discrepancies" +
    " in some areas.\n" +
    "A rating of 0 means that the User Answer is either " +
    "inaccurate, incomplete, or unrelated to the Reference " +
    "Answer, or it fails to address the Question.\n" +
    "I will provide the rating without any explanation or " +
    "justification, adhering to the following scale: " +
    "0 (no match), 2 (partial match), 4 (exact match).\n" +
    "Do not explain or justify my rating. My rating must" +
    " be only 4, 2 or 0 only.\n\n" + "Question: \"{question}\"\n\n" +
    "Reference Answer: \"{model_pred}\"\n\n" +
    "User Answer: \"{correct_answer}\"\n\n" + "Rating: ")


# TODO: add support for Local LM
# TODO: add multiproc support like txt2kg
class LLMJudge():
    """Uses NIMs to score a triple of (question, model_pred, correct_answer)
    This whole class is an adaptation of Gilberto's work for PyG.

    Args:
        NVIDIA_NIM_MODEL : (str, optional)
            The name of the NVIDIA NIM model to use.
            (default: "nvidia/llama-3.1-nemotron-70b-instruct").
        NVIDIA_API_KEY : (str, optional)
            The API key for accessing NVIDIA's NIM models.
            (default: "").
        ENDPOINT_URL : (str, optional)
            The URL hosting your model, in case you are not using
            the public NIM.
            (default: "https://integrate.api.nvidia.com/v1").
    """
    def __init__(
        self,
        NVIDIA_NIM_MODEL: Optional[
            str] = "nvidia/llama-3.1-nemotron-70b-instruct",
        NVIDIA_API_KEY: Optional[str] = "",
        ENDPOINT_URL: Optional[str] = "https://integrate.api.nvidia.com/v1",
    ) -> None:
        self.NVIDIA_API_KEY = NVIDIA_API_KEY
        self.NIM_MODEL = NVIDIA_NIM_MODEL
        self.ENDPOINT_URL = ENDPOINT_URL

    def _process_score(self, response: str) -> float:
        """Uses 3 and 1 even though prompt says only 0, 2, 4.
        This is because LLMs don't always follow instructions.
        Credit to Gilberto.
        """
        for i in [4, 3, 2, 1, 0]:
            if str(i) in response:
                return i / 4
        return float("nan")

    def _average_scores(self, score0: float, score1: float):
        """Take the average of score0 and score1.
        Sometimes the LLM fail to respond or have no score in the response.
        In those cases the failed score is discarded.
        Credit to Gilberto.

        Args:
         score0 (float): judge accuracy score.
         score1 (float): judge accuracy score by permuting agent answer and
         ground truth.

        Returns:
            (float) average of score0 and score1 of both contains scores,
            otherwise pick the max.
        """
        score = float("nan")
        if score0 >= 0 and score1 >= 0:
            score = (score0 + score1) / 2
        else:
            score = max(score0, score1)
        return score

    def score(
        self,
        question: str,
        model_pred: str,
        correct_answer: str,
    ) -> float:
        """Args:
            question (str): The original question asked to the model.
            model_pred (str): The prediction made by the model.
            correct_answer (str): The actual correct answer to the question.

        Returns:
            score (float): score of 0-1, may be nan due to LLM judge failure.
                Evals should skip nan's when aggregating score.
        """
        prompt1 = SYSTEM_PROMPT_1.format(question=question,
                                         model_pred=model_pred,
                                         correct_answer=correct_answer)
        prompt2 = SYSTEM_PROMPT_2.format(question=question,
                                         model_pred=model_pred,
                                         correct_answer=correct_answer)
        score1 = float("nan")
        score2 = float("nan")
        for _retry in range(200):
            try:
                score1 = self._process_score(
                    call_NIM(prompt1, self.NVIDIA_API_KEY, self.NIM_MODEL,
                             self.ENDPOINT_URL, post_text=""))
                if not isnan(score1):
                    break
            except ImportError:
                raise
            except:  # noqa
                pass
        for _retry in range(20):
            try:
                score2 = self._process_score(
                    call_NIM(prompt2, self.NVIDIA_API_KEY, self.NIM_MODEL,
                             self.ENDPOINT_URL, post_text=""))
                if not isnan(score2):
                    break
            except ImportError:
                raise
            except:  # noqa
                pass

        return self._average_scores(score1, score2)



from typing import Optional, Tuple

import torch

from torch_geometric.data import Data
from torch_geometric.data.datapipes import functional_transform
from torch_geometric.transforms import BaseTransform


@functional_transform('distance')
class Distance(BaseTransform):
    r"""Saves the Euclidean distance of linked nodes in its edge attributes
    (functional name: :obj:`distance`). Each distance gets globally normalized
    to a specified interval (:math:`[0, 1]` by default).

    Args:
        norm (bool, optional): If set to :obj:`False`, the output will not be
            normalized. (default: :obj:`True`)
        max_value (float, optional): If set and :obj:`norm=True`, normalization
            will be performed based on this value instead of the maximum value
            found in the data. (default: :obj:`None`)
        cat (bool, optional): If set to :obj:`False`, all existing edge
            attributes will be replaced. (default: :obj:`True`)
        interval ((float, float), optional): A tuple specifying the lower and
            upper bound for normalization. (default: :obj:`(0.0, 1.0)`)
    """
    def __init__(
            self,
            norm: bool = True,
            max_value: Optional[float] = None,
            cat: bool = True,
            interval: Tuple[float, float] = (0.0, 1.0),
    ):
        self.norm = norm
        self.max = max_value
        self.cat = cat
        self.interval = interval

    def forward(self, data: Data) -> Data:
        assert data.pos is not None
        assert data.edge_index is not None
        (row, col), pos, pseudo = data.edge_index, data.pos, data.edge_attr

        dist = torch.norm(pos[col] - pos[row], p=2, dim=-1).view(-1, 1)

        if self.norm and dist.numel() > 0:
            max_val = float(dist.max()) if self.max is None else self.max

            length = self.interval[1] - self.interval[0]
            dist = length * (dist / max_val) + self.interval[0]

        if pseudo is not None and self.cat:
            pseudo = pseudo.view(-1, 1) if pseudo.dim() == 1 else pseudo
            data.edge_attr = torch.cat([pseudo, dist.type_as(pseudo)], dim=-1)
        else:
            data.edge_attr = dist

        return data

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}(norm={self.norm}, '
                f'max_value={self.max})')



import random
from typing import Tuple

from torch_geometric.data import Data
from torch_geometric.data.datapipes import functional_transform
from torch_geometric.transforms import BaseTransform


@functional_transform('random_scale')
class RandomScale(BaseTransform):
    r"""Scales node positions by a randomly sampled factor :math:`s` within a
    given interval, *e.g.*, resulting in the transformation matrix
    (functional name: :obj:`random_scale`).

    .. math::
        \begin{bmatrix}
            s & 0 & 0 \\
            0 & s & 0 \\
            0 & 0 & s \\
        \end{bmatrix}

    for three-dimensional positions.

    Args:
        scales (tuple): scaling factor interval, e.g. :obj:`(a, b)`, then scale
            is randomly sampled from the range
            :math:`a \leq \mathrm{scale} \leq b`.
    """
    def __init__(self, scales: Tuple[float, float]) -> None:
        assert isinstance(scales, (tuple, list)) and len(scales) == 2
        self.scales = scales

    def forward(self, data: Data) -> Data:
        assert data.pos is not None

        scale = random.uniform(*self.scales)
        data.pos = data.pos * scale
        return data

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.scales})'
